#ifndef BOOST_SERIALIZER_OPS
#define BOOST_SERIALIZER_OPS

#include <boost/type_traits.hpp>
#include <functional>

namespace boost {

namespace serializer {

// Cast op
// Returns whatever is passed to act as a "dead end" of cast ops
template <typename Type>
struct nop : public std::unary_function<typename boost::add_reference<typename boost::add_const<Type>::type>::type,
                                        Type> {
	typedef nop <Type> this_type;

	typename this_type::result_type operator()(typename this_type::argument_type v) const {
		return v;
	}
};

// Wrapper pack op
// Used to take a pack op and apply it over a standard library style container
template <typename RecvType, typename ContainerType, template <typename, typename, typename> class PackOp, typename CastOp = nop <typename ContainerType::value_type> >
struct for_each : std::binary_function<typename boost::add_reference<RecvType>::type,
                                       typename boost::add_reference<typename boost::add_const<ContainerType>::type>::type,
                                       void> {
	typedef for_each <RecvType, ContainerType, PackOp, CastOp>                 this_type;
	typedef RecvType                                                           receive_type;
	typedef ContainerType                                                      container_type;
	typedef PackOp <receive_type, typename container_type::value_type, CastOp> pack_op_type;

	pack_op_type pack_op;

	void operator()(typename this_type::first_argument_type receive, typename this_type::second_argument_type v) const {
		typename container_type::const_iterator cur, end = v.end();

		for(cur = v.begin(); cur != end; ++cur)
			pack_op(receive, *cur);
	}

	for_each(const pack_op_type &op = pack_op_type()) : pack_op(op) {
	}
};

// Pack op
// Only calls the cast op
template <typename RecvType, typename Type, typename CastOp = nop <Type> >
struct call : std::binary_function<typename boost::add_reference<RecvType>::type,
                                   typename boost::add_reference<typename boost::add_const<Type>::type>::type,
                                   void> {
	typedef call <RecvType, Type, CastOp> this_type;
	typedef RecvType                      receive_type;
	typedef Type                          type_type;
	typedef CastOp                        cast_op_type;

	cast_op_type cast_op;

	void operator()(typename this_type::first_argument_type receive, typename this_type::second_argument_type v) const {
		receive(cast_op(v));
	}

	call(const cast_op_type &op = cast_op_type()) : cast_op(op) {
	}
};

// Pack op
// Calls a custom functor and helpfully eats the normally helpful template arguments
template <typename FuncType>
struct custom {
	template <typename RecvType, typename Type, typename CastOp = nop <Type> >
	struct type : std::binary_function<typename boost::add_reference<RecvType>::type,
                                           typename boost::add_reference<typename boost::add_const<Type>::type>::type,
                                           void> {
		typedef type <RecvType, Type, CastOp> this_type;
		typedef FuncType                      func_type;

		func_type func;

		void operator()(typename this_type::first_argument_type receive, typename this_type::second_argument_type v) const {
			func(receive, v);
		}

		type(const func_type &f = func_type()) : func(f) {
		}
	};
};

// Pack op
// Uses the insertion operator (left shift) to add items to a receiver
template <typename RecvType, typename Type, typename CastOp = nop <Type> >
struct insert : std::binary_function<typename boost::add_reference<RecvType>::type,
                                     typename boost::add_reference<typename boost::add_const<Type>::type>::type,
                                     void> {
	typedef insert <RecvType, Type, CastOp> this_type;
	typedef RecvType                        receive_type;
	typedef Type                            type_type;
	typedef CastOp                          cast_op_type;

	cast_op_type cast_op;

	void operator()(typename this_type::first_argument_type receive, typename this_type::second_argument_type v) const {
		receive << cast_op(v);
	}

	insert(const cast_op_type &op = cast_op_type()) : cast_op(op) {
	}
};

// Pack op
// Only calls the cast op
template <typename RecvType, typename Type, typename CastOp = nop <Type> >
struct nothing : std::binary_function<typename boost::add_reference<RecvType>::type,
                                      typename boost::add_reference<typename boost::add_const<Type>::type>::type,
                                      void> {
	typedef nothing <RecvType, Type, CastOp> this_type;
	typedef RecvType                         receive_type;
	typedef Type                             type_type;
	typedef CastOp                           cast_op_type;

	cast_op_type cast_op;

	void operator()(typename this_type::first_argument_type receive, typename this_type::second_argument_type v) const {
		cast_op(v);
	}

	nothing(const cast_op_type &op = cast_op_type()) : cast_op(op) {
	}
};

}

}

#endif
